from dune.perftool.generation import (backend,
                                      temporary_variable,
                                      instruction)
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.options import get_option
from dune.perftool.pdelab.geometry import (name_jacobian_determinant,
                                           name_jacobian_inverse_transposed,
                                           local_dimension,
                                           apply_to_global_transformation,
                                           name_facet_jacobian_determinant)
from dune.perftool.blockstructured.tools import sub_element_inames
import pymbolic.primitives as prim


@backend(interface="pymbolic_jacobian_determinant", name="blockstructured")
def pymbolic_jacobian_determinant():
    return prim.Quotient(prim.Variable(name_jacobian_determinant()),
                         prim.Power(get_option("number_of_blocks"),local_dimension()))


@backend(interface="pymbolic_jacobian_inverse_transposed", name="blockstructured")
def pymbolic_jacobian_inverse_transposed(i,j,restriction):
    return prim.Product((get_option("number_of_blocks"),
                         prim.Subscript(prim.Variable(name_jacobian_inverse_transposed(restriction)), (j, i))))


@backend(interface="pymbolic_facet_jacobian_determinant", name="blockstructured")
def pymbolic_facet_jacobian_determinant():
    return prim.Quotient(prim.Variable(name_facet_jacobian_determinant()),
                         prim.Power(get_option("number_of_blocks"),local_dimension()))


def define_point_in_macro(name, point_in_micro):
    dim = local_dimension()
    temporary_variable(name, shape_impl=('fv',), shape=(dim,))

    subelem_inames = sub_element_inames()
    for i in range(dim):
        if isinstance(point_in_micro, prim.Subscript):
            expr = prim.Subscript(point_in_micro.aggregate, point_in_micro.index + (i,))
        else:
            expr = prim.Subscript(point_in_micro, (i,))
        expr = prim.Sum((expr, prim.Variable(subelem_inames[i]),))
        expr = prim.Quotient(expr, get_option('number_of_blocks'))
        instruction(assignee=prim.Subscript(prim.Variable(name), (i,)),
                    expression=expr,
                    within_inames=frozenset(subelem_inames),
                    )


def name_point_in_macro(point_in_micro):
    assert isinstance(point_in_micro, prim.Expression)
    name = get_pymbolic_basename(point_in_micro) + "_macro"
    define_point_in_macro(name, point_in_micro)
    return name


@backend(interface="to_global", name="blockstructured")
def to_global(local):
    macro = name_point_in_macro(local)
    name = macro + "_global"
    apply_to_global_transformation(name, prim.Variable(macro))
    return prim.Variable(name)
