from dune.perftool.ufl.modified_terminals import Restriction
from dune.perftool.generation import (iname,
                                      domain,
                                      get_global_context_value,
                                      temporary_variable,
                                      instruction)
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           name_localcenter,
                                           pymbolic_in_cell_coordinates)

from dune.perftool.pdelab.quadrature import (quadrature_inames)
from dune.perftool.generation.counter import get_counted_variable
from dune.perftool.options import get_option
import pymbolic.primitives as prim


@iname
def sub_element_inames():
    name = "subel"
    dim = local_dimension()
    inames = tuple()
    for i in range(dim):
        name_counted = get_counted_variable(name)
        inames = inames + (name_counted,)
        domain(name_counted, get_option("number_of_blocks"))
    return inames


def tensor_index_to_sequential_index(inames, k):
    index = prim.Sum(tuple(prim.Variable(name)*k**i for i,name in enumerate(inames)))
    return index


# TODO better name
def to_tensor_index(iname):
    return tuple(prim.RightShift(prim.BitwiseAnd((prim.Variable(iname),2**n)), n)  for n in range(world_dimension()))


# TODO 3d
def micro_index_to_macro_index(iname):

    subelem_inames_orig = sub_element_inames()

    k = get_option("number_of_blocks")

    it = get_global_context_value("integral_type")
    if it == "cell":
        sublem_inames = subelem_inames_orig
    elif it == "exterior_facet" or it == "interior_facet":
        center = pymbolic_in_cell_coordinates(prim.Variable(name_localcenter()), Restriction.NEGATIVE)

        def predicate(x):
            return frozenset([prim.Comparison(prim.Subscript(center, 0), x, 0.5)])

        sublem_inames = ("x",)
        temporary_variable(sublem_inames[0])
        instruction(assignee=prim.Variable(sublem_inames[0]),
                    expression=prim.Variable(subelem_inames_orig[0]),
                    within_inames=frozenset(subelem_inames_orig).union(frozenset(quadrature_inames())),
                    predicates=predicate("==")
                    )
        instruction(assignee=prim.Variable(sublem_inames[0]),
                    expression=prim.Product(((k-1),prim.Subscript(center, (0,)))),
                    within_inames=frozenset(subelem_inames_orig).union(frozenset(quadrature_inames())),
                    predicates=predicate("!=")
                    )
        sublem_inames = sublem_inames + ("y",)
        temporary_variable(sublem_inames[1])
        instruction(assignee=prim.Variable(sublem_inames[1]),
                    expression=prim.Product(((k-1),prim.Subscript(center, (1,)))),
                    within_inames=frozenset(subelem_inames_orig).union(frozenset(quadrature_inames())),
                    predicates=predicate("==")
                    )
        instruction(assignee=prim.Variable(sublem_inames[1]),
                    expression=prim.Variable(subelem_inames_orig[0]),
                    within_inames=frozenset(subelem_inames_orig).union(frozenset(quadrature_inames())),
                    predicates=predicate("!=")
                    )
    #
    # modified_index = prim.Sum((prim.Variable(x), index))
    # if local_dimension() > 1:
    #     modified_index = prim.Sum((modified_index, prim.Product(((k+1), prim.Variable(y)))))
    #     index_div_2 = prim.FloorDiv(index, 2)
    #     index_div_2 = prim.Product((index_div_2, k-1))
    #     modified_index = prim.Sum((modified_index, index_div_2))
    #     if local_dimension() > 2:
    #         modified_index = prim.Sum((modified_index, prim.Variable(z)))

    modified_index = prim.Sum(tuple(to_tensor_index(iname)[n] * (k + 1) ** n for n in range(world_dimension()))
                              + (tensor_index_to_sequential_index(sublem_inames, get_option("number_of_blocks") + 1), ))

    return modified_index

